package de.ozgcloud.xta.test.app.data;

import java.io.IOException;
import java.math.BigInteger;
import java.nio.file.Files;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;

import org.springframework.stereotype.Component;

import de.ozgcloud.xta.test.app.model.XtaFile;
import de.ozgcloud.xta.test.app.model.XtaMessage;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Component
@Log4j2
@RequiredArgsConstructor
public class XtaMessageRepository {

	private final Map<String, XtaMessage> messages = new HashMap<>();

	public void save(XtaMessage message) {
		messages.put(message.metaData().messageId(), message);
	}

	public Optional<XtaMessage> get(String messageId) {
		return Optional.ofNullable(messages.get(messageId));
	}

	public boolean delete(String messageId) {
		if (messages.containsKey(messageId)) {
			var message = messages.get(messageId);
			message.attachmentFiles().forEach(this::silentDeleteFile);
			silentDeleteFile(message.messageFile());
			messages.remove(messageId);
			return true;
		}
		return false;
	}

	private void silentDeleteFile(XtaFile xtaFile) {
		try {
			Files.delete(xtaFile.file().toPath());
		} catch (IOException e) {
			log.warn("Could not delete file: {}", e.getMessage());
		}
	}

	public List<XtaMessage> findByReaderIdentifierValueLimitedTo(String readerIdentifierValue, int maxListItems) {
		return findByReaderIdentifierValue(readerIdentifierValue)
				.limit(maxListItems)
				.toList();
	}

	public BigInteger countByReaderIdentifierValue(String readerIdentifierValue) {
		return BigInteger.valueOf(findByReaderIdentifierValue(readerIdentifierValue).count());
	}

	Stream<XtaMessage> findByReaderIdentifierValue(String readerIdentifierValue) {
		return messages.values().stream()
				.filter(m -> m.metaData().readerIdentifier().value().equals(readerIdentifierValue));
	}

	public boolean deleteByMessageIdAndReaderIdentifierValue(String messageId, String readerIdentifierValue) {
		return get(messageId)
				.filter(m -> m.metaData().readerIdentifier().value().equals(readerIdentifierValue))
				.map(m -> delete(messageId))
				.orElse(false);
	}
}
