package de.ozgcloud.xta.test.app.data;

import static de.ozgcloud.xta.test.app.factory.XtaMessageMetaDataTestFactory.*;
import static de.ozgcloud.xta.test.app.factory.XtaMessageTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Spy;

import de.ozgcloud.xta.test.app.factory.XtaFileTestFactory;
import de.ozgcloud.xta.test.app.factory.XtaMessageMetaDataTestFactory;
import de.ozgcloud.xta.test.app.factory.XtaMessageTestFactory;
import de.ozgcloud.xta.test.app.model.XtaFile;
import de.ozgcloud.xta.test.app.model.XtaIdentifier;
import de.ozgcloud.xta.test.app.model.XtaMessage;

public class XtaMessageRepositoryTest {

	@Spy
	@InjectMocks
	private XtaMessageRepository xtaMessageRepository;

	@DisplayName("save")
	@Nested
	class TestSave {

		@Test
		void shouldSaveMessage() {
			var message = XtaMessageTestFactory.create(MESSAGE_ID_1);

			xtaMessageRepository.save(message);

			assertThat(xtaMessageRepository.get(MESSAGE_ID_1)).contains(message);
		}
	}

	@DisplayName("get")
	@Nested
	class TestGet {
		private final XtaMessage message = XtaMessageTestFactory.create(MESSAGE_ID_1);

		@BeforeEach
		void beforeEach() {
			xtaMessageRepository = new XtaMessageRepository();
			xtaMessageRepository.save(message);
		}

		@Test
		void shouldReturnMessage() {
			var result = xtaMessageRepository.get(MESSAGE_ID_1);

			assertThat(result).contains(message);
		}

		@Test
		void shouldReturnEmpty() {
			var result = xtaMessageRepository.get(MESSAGE_ID_2);

			assertThat(result).isEmpty();
		}
	}

	@Nested
	class TestDelete {
		public static final String MESSAGE_ID = "messageId";
		private XtaFile messageFile;
		private XtaFile attachmentFile;

		@BeforeEach
		void beforeEach() {
			messageFile = XtaFileTestFactory.create();
			attachmentFile = XtaFileTestFactory.create();
			var message = XtaMessageTestFactory.createBuilder(MESSAGE_ID)
					.messageFile(messageFile)
					.attachmentFiles(List.of(attachmentFile))
					.build();
			xtaMessageRepository.save(message);
		}

		@Test
		void shouldDeleteMessage() {
			xtaMessageRepository.delete(MESSAGE_ID);

			assertThat(xtaMessageRepository.get(MESSAGE_ID)).isEmpty();
		}

		@Test
		void shouldDeleteTemporaryFiles() {
			assertThat(messageFile.file()).exists();
			assertThat(attachmentFile.file()).exists();

			xtaMessageRepository.delete(MESSAGE_ID);

			assertThat(messageFile.file()).doesNotExist();
			assertThat(attachmentFile.file()).doesNotExist();
		}
	}

	@DisplayName("find by reader identifier value limited to")
	@Nested
	class TestFindByReaderIdentifierValueLimitedTo {

		@BeforeEach
		void beforeEach() {
			xtaMessageRepository.save(XtaMessageTestFactory.create("1"));
			xtaMessageRepository.save(XtaMessageTestFactory.create("2"));
			xtaMessageRepository.save(XtaMessageTestFactory.create("3"));
		}

		@DisplayName("should return limit")
		@ParameterizedTest
		@ValueSource(ints = { 0, 1, 2 })
		void shouldReturnLimit(int limit) {
			var result = xtaMessageRepository.findByReaderIdentifierValueLimitedTo(READER_IDENTIFIER.value(), limit);

			assertThat(result).hasSize(limit);
		}

		@DisplayName("should return all")
		@ParameterizedTest
		@ValueSource(ints = { 3, 4, 5 })
		void shouldReturnAll(int limit) {
			var result = xtaMessageRepository.findByReaderIdentifierValueLimitedTo(READER_IDENTIFIER.value(), limit);

			assertThat(result).hasSize(3);
		}

		@DisplayName("should not find any with unknown reader identifier")
		@Test
		void shouldNotFindAnyWithUnknownReaderIdentifier() {
			var result = xtaMessageRepository.findByReaderIdentifierValueLimitedTo("unknown", 1);

			assertThat(result).isEmpty();
		}

	}

	@DisplayName("count by reader identifier value")
	@Nested
	class TestCountByReaderIdentifierValue {
		@BeforeEach
		void beforeEach() {
			doReturn(Stream.of(XtaMessageTestFactory.create("1"), XtaMessageTestFactory.create("2"),
					XtaMessageTestFactory.create("3"))).when(xtaMessageRepository).findByReaderIdentifierValue(READER_IDENTIFIER.value());
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var count = xtaMessageRepository.countByReaderIdentifierValue(READER_IDENTIFIER.value());

			assertThat(count).isEqualTo(3);
		}

	}

	@DisplayName("find by reader identifier value")
	@Nested
	class TestFindByReaderIdentifierValue {
		private static final String OTHER_READER_IDENTIFIER = "readerIdentifier";

		@BeforeEach
		void beforeEach() {
			xtaMessageRepository = new XtaMessageRepository();
			xtaMessageRepository.save(XtaMessageTestFactory.create("1"));
			xtaMessageRepository.save(XtaMessageTestFactory.create("2"));
			xtaMessageRepository.save(XtaMessageTestFactory.create("3"));
			xtaMessageRepository.save(XtaMessageTestFactory.createBuilder()
					.metaData(XtaMessageMetaDataTestFactory.createBuilder()
							.readerIdentifier(XtaIdentifier.builder().value(OTHER_READER_IDENTIFIER).build())
							.build())
					.build());
		}

		@DisplayName("should return message for reader identifier")
		@Test
		void shouldReturnMessageForReaderIdentifier() {
			var result = xtaMessageRepository.findByReaderIdentifierValue(READER_IDENTIFIER.value()).toList();

			assertThat(result).hasSize(3);
		}

		@DisplayName("should return messages for other reader identifier")
		@Test
		void shouldReturnMessagesForOtherReaderIdentifier() {
			var result = xtaMessageRepository.findByReaderIdentifierValue(OTHER_READER_IDENTIFIER).toList();

			assertThat(result).hasSize(1);
		}

		@DisplayName("should return no messages for unknown reader identifier")
		@Test
		void shouldReturnNoMessagesForUnknownReaderIdentifier() {
			var result = xtaMessageRepository.findByReaderIdentifierValue("unknown").toList();

			assertThat(result).isEmpty();
		}

	}
}
