package de.ozgcloud.xta.test.app.service;

import static de.ozgcloud.xta.test.app.factory.XtaMessageMetaDataTestFactory.*;
import static de.ozgcloud.xta.test.app.factory.XtaMessageTestFactory.*;
import static de.ozgcloud.xta.test.app.validation.ValidXtaId.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.math.BigInteger;
import java.util.List;
import java.util.Optional;
import java.util.function.Predicate;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.core.env.Environment;

import de.ozgcloud.xta.test.app.data.XtaMessageRepository;
import de.ozgcloud.xta.test.app.data.XtaTransportReportRepository;
import de.ozgcloud.xta.test.app.factory.XtaCloseRequestTestFactory;
import de.ozgcloud.xta.test.app.factory.XtaMessageMetaDataTestFactory;
import de.ozgcloud.xta.test.app.factory.XtaMessageTestFactory;
import de.ozgcloud.xta.test.app.model.*;

public class XtaMessageServiceTest {
	@Mock
	private XtaMessageRepository messageRepository;
	@Mock
	private XtaTransportReportRepository transportReportRepository;
	@Mock
	private Environment environment;

	@Spy
	@InjectMocks
	private XtaMessageService service;

	@Nested
	class TestSendMessage {
		public static final String MESSAGE_ID = "send_1";
		private XtaMessage message;

		@BeforeEach
		void beforeEach() {
			message = XtaMessageTestFactory.create(MESSAGE_ID);
		}

		@Test
		void shouldSaveTransportReport() {
			message = XtaMessageTestFactory.create(MESSAGE_ID);

			service.sendMessage(message);

			verify(transportReportRepository).add(message.metaData());
		}

		@Test
		void shouldSaveMessage() {
			when(transportReportRepository.add(message.metaData())).thenReturn(true);
			message = XtaMessageTestFactory.create(MESSAGE_ID);

			service.sendMessage(message);

			verify(messageRepository).save(message);
		}
	}

	@Nested
	class TestGetStatusList {

		private static final int MAX_LIST_ITEMS = 50;
		private static final int MESSAGE_COUNT = 2;

		private final XtaMessageMetaDataListingRequest messageMetaDataListingRequest = XtaMessageMetaDataListingRequest.builder()
				.clientIdentifier(READER_IDENTIFIER)
				.maxListItems(MAX_LIST_ITEMS)
				.build();

		private XtaMessage message1;
		private XtaMessage message2;

		@BeforeEach
		void beforeEach() {
			message1 = XtaMessageTestFactory.create(MESSAGE_ID_1);
			message2 = XtaMessageTestFactory.create(MESSAGE_ID_2);

			when(messageRepository.findByReaderIdentifierValueLimitedTo(READER_IDENTIFIER.value(), MAX_LIST_ITEMS)).thenReturn(
					List.of(message1, message2));
			when(messageRepository.countByReaderIdentifierValue(READER_IDENTIFIER.value())).thenReturn(BigInteger.valueOf(MESSAGE_COUNT));
		}

		@Test
		void shouldGetMessageMetadataFromRepository() {
			var statusList = getStatusList();

			assertThat(statusList.messages()).containsOnly(message1.metaData(), message2.metaData());
		}

		@Test
		void shouldSetPendingMessageToMessageCount() {
			var statusList = getStatusList();

			assertThat(statusList.pendingMessageCount()).isEqualTo(MESSAGE_COUNT);
		}

		private XtaMessageMetaDataListing getStatusList() {
			return service.getStatusList(messageMetaDataListingRequest);
		}

	}

	@DisplayName("get message")
	@Nested
	class TestGetMessage {
		private static final XtaGetMessageRequest GET_MESSAGE_REQUEST = XtaGetMessageRequest.builder()
				.messageId(MESSAGE_ID_1)
				.clientIdentifier(READER_IDENTIFIER)
				.build();
		private static final Predicate<XtaMessage> IS_AUTHOR_OR_READER = m -> true;
		private static final Predicate<XtaMessage> IS_NOT_AUTHOR_OR_READER = m -> false;

		@Mock
		private XtaMessage message;

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			when(messageRepository.get(MESSAGE_ID_1)).thenReturn(Optional.of(message));
			doReturn(IS_AUTHOR_OR_READER).when(service).isAuthorOrReaderOfMessage(READER_IDENTIFIER.value());

			var result = service.getMessage(GET_MESSAGE_REQUEST);

			assertThat(result).contains(message);
		}

		@DisplayName("should return empty if message id not found")
		@Test
		void shouldReturnEmptyIfMessageIdNotFound() {
			when(messageRepository.get(MESSAGE_ID_1)).thenReturn(Optional.empty());

			var result = service.getMessage(GET_MESSAGE_REQUEST);

			assertThat(result).isEmpty();
		}

		@DisplayName("should return empty if not author or reader")
		@Test
		void shouldReturnEmptyIfNotAuthorOrReader() {
			when(messageRepository.get(MESSAGE_ID_1)).thenReturn(Optional.of(message));
			doReturn(IS_NOT_AUTHOR_OR_READER).when(service).isAuthorOrReaderOfMessage(READER_IDENTIFIER.value());

			var result = service.getMessage(GET_MESSAGE_REQUEST);

			assertThat(result).isEmpty();
		}
	}

	@DisplayName("is author or reader of message")
	@Nested
	class TestIsAuthorOrReaderOfMessage {
		@Test
		void shouldUseIsAuthorOrReaderOfMessageMetaData() {
			var message = XtaMessageTestFactory.create(MESSAGE_ID_1);

			service.isAuthorOrReaderOfMessage(READER_IDENTIFIER.value()).test(message);

			verify(service).isAuthorOrReaderOfMessageMetadata(READER_IDENTIFIER.value());
		}
	}

	@DisplayName("is author or reader of transport report")
	@Nested
	class TestIsAuthorOrReaderOfTransportReport {
		@Test
		void shouldUseIsAuthorOrReaderOfMessageMetaData() {
			var transportReport = XtaTransportReport.builder()
					.messageMetaData(XtaMessageMetaDataTestFactory.create(MESSAGE_ID_1))
					.build();

			service.isAuthorOrReaderOfTransportReport(READER_IDENTIFIER.value()).test(transportReport);

			verify(service).isAuthorOrReaderOfMessageMetadata(READER_IDENTIFIER.value());
		}
	}

	@DisplayName("is author or reader of message metadata")
	@Nested
	class TestIsAuthorOrReaderOfMessageMetaData {
		@DisplayName("should return true if reader identifier value matches")
		@Test
		void shouldReturnTrueIfReaderIdentifierValueMatches() {
			var messageMetaData = XtaMessageMetaDataTestFactory.create(MESSAGE_ID_1);

			var result = service.isAuthorOrReaderOfMessageMetadata(READER_IDENTIFIER.value()).test(messageMetaData);

			assertThat(result).isTrue();
		}

		@DisplayName("should return true if author identifier value matches")
		@Test
		void shouldReturnTrueIfAuthorIdentifierValueMatches() {
			var messageMetaData = XtaMessageMetaDataTestFactory.create(MESSAGE_ID_1);

			var result = service.isAuthorOrReaderOfMessageMetadata(AUTHOR_IDENTIFIER.value()).test(messageMetaData);

			assertThat(result).isTrue();
		}

		@DisplayName("should return false if neither reader nor author identifier value matches")
		@Test
		void shouldReturnFalseIfNeitherReaderNorAuthorIdentifierValueMatches() {
			var messageMetaData = XtaMessageMetaDataTestFactory.create(MESSAGE_ID_1);

			var result = service.isAuthorOrReaderOfMessageMetadata("unknown").test(messageMetaData);

			assertThat(result).isFalse();
		}
	}

	@DisplayName("close message")
	@Nested
	class TestCloseMessage {

		@DisplayName("should return")
		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		void shouldReturn(boolean deleted) {
			var request = XtaCloseRequestTestFactory.create();
			when(messageRepository.deleteByMessageIdAndReaderIdentifierValue(
					request.messageId(), request.clientIdentifier().value())
			).thenReturn(deleted);
			when(transportReportRepository.setClosedStatus(request.messageId())).thenReturn(deleted);

			boolean result = service.closeMessage(request);

			assertThat(result).isEqualTo(deleted);
		}

		@DisplayName("should set transport status to closed")
		@Test
		void shouldSetTransportReportClosedStatus() {
			var request = XtaCloseRequestTestFactory.create();

			service.closeMessage(request);

			verify(transportReportRepository).setClosedStatus(request.messageId());
		}
	}

	@Nested
	class TestGetTransportReport {
		private static final XtaGetTransportReportRequest GET_TRANSPORT_REPORT_REQUEST = XtaGetTransportReportRequest.builder()
				.messageId(MESSAGE_ID_1)
				.clientIdentifier(READER_IDENTIFIER)
				.build();
		private static final Predicate<XtaTransportReport> IS_AUTHOR_OR_READER = m -> true;
		private static final Predicate<XtaTransportReport> IS_NOT_AUTHOR_OR_READER = m -> false;

		@Mock
		XtaTransportReport report;

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			when(transportReportRepository.get(MESSAGE_ID_1)).thenReturn(Optional.of(report));
			doReturn(IS_AUTHOR_OR_READER).when(service).isAuthorOrReaderOfTransportReport(READER_IDENTIFIER.value());

			var result = service.getTransportReport(GET_TRANSPORT_REPORT_REQUEST);

			assertThat(result).contains(report);
		}

		@DisplayName("should return empty if message id not found")
		@Test
		void shouldReturnEmptyIfMessageIdNotFound() {
			when(transportReportRepository.get(MESSAGE_ID_1)).thenReturn(Optional.empty());

			var result = service.getTransportReport(GET_TRANSPORT_REPORT_REQUEST);

			assertThat(result).isEmpty();
		}

		@DisplayName("should return empty if not author or reader")
		@Test
		void shouldReturnEmptyIfNotAuthorOrReader() {
			when(transportReportRepository.get(MESSAGE_ID_1)).thenReturn(Optional.of(report));
			doReturn(IS_NOT_AUTHOR_OR_READER).when(service).isAuthorOrReaderOfTransportReport(READER_IDENTIFIER.value());

			var result = service.getTransportReport(GET_TRANSPORT_REPORT_REQUEST);

			assertThat(result).isEmpty();
		}
	}

	@DisplayName("create message id")
	@Nested
	class TestCreateMessageId {

		@DisplayName("should return unique id")
		@Test
		void shouldReturnUniqueId() {
			var result1 = service.createMessageId();
			var result2 = service.createMessageId();

			assertThat(result1).isNotEqualTo(result2);
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			when(environment.getProperty("spring.application.name")).thenReturn("app");

			var result = service.createMessageId();

			assertThat(result).startsWith("urn:" + MESSAGE_ID_NAMESPACE + ":app:");
		}
	}

	@DisplayName("create request id")
	@Nested
	class TestCreateRequestId {

		@DisplayName("should return unique id")
		@Test
		void shouldReturnUniqueId() {
			var result1 = service.createRequestId();
			var result2 = service.createRequestId();

			assertThat(result1).isNotEqualTo(result2);
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			when(environment.getProperty("spring.application.name")).thenReturn("app");

			var result = service.createRequestId();

			assertThat(result).startsWith("urn:" + REQUEST_ID_NAMESPACE + ":app:");
		}
	}

}
